﻿using System;
using System.Text;
using EonaCat.LogSystem;
using EonaCat.Quic;
using EonaCat.Quic.Connections;
using EonaCat.Quic.Events;
using EonaCat.Quic.Helpers;
using EonaCat.Quic.Streams;

namespace EonaCat.Network
{
    // This file is part of the EonaCat project(s) which is released under the Apache License.
    // See the LICENSE file or go to https://EonaCat.com/License for full license details.

    public static class NetworkHelper
    {
        internal static Logging Logger = new Logging();
        private static QuicServer _quicServer;

        public static event EventHandler<QuicConnectionEventArgs> OnQuicClientConnected;

        public static event EventHandler<QuicStreamEventArgs> OnQuicStreamOpened;

        public static event EventHandler<QuicStreamEventArgs> OnQuicStreamDataReceived;

        /// <summary>
        /// Start a Quic server
        /// </summary>
        /// <returns><c>true</c>, start successfully, <c>false</c> did not start successfully.</returns>
        /// <param name="ip">The ip bound to the server</param>
        /// <param name="port">The listening port. (default: 11000)</param>
        public static bool QuicStartServer(string ip, int port = 11000)
        {
            _quicServer = new QuicServer(ip, port);
            _quicServer.OnClientConnected += ClientConnected;
            _quicServer.Start();
            Logger.Info($"The Quic server has been successfully started on ip '{ip}' and port: {port}");
            return true;
        }

        /// <summary>
        /// Fired when Client is connected
        /// </summary>
        /// <param name="connection">The new connection</param>
        private static void ClientConnected(QuicConnection connection)
        {
            OnQuicClientConnected?.Invoke(null, new QuicConnectionEventArgs { Connection = connection });
            connection.OnStreamOpened += StreamOpened;
        }

        private static void StreamOpened(QuicStream stream)
        {
            OnQuicStreamOpened?.Invoke(null, new QuicStreamEventArgs { Stream = stream });
            stream.OnStreamDataReceived += StreamDataReceived;
        }

        private static void StreamDataReceived(QuicStream stream, byte[] data)
        {
            OnQuicStreamDataReceived?.Invoke(null, new QuicStreamEventArgs { Stream = stream, Data = data });
        }

        public static QuicStream QuicStartClient(string ip, int port = 11000, StreamType streamType = StreamType.ClientBidirectional)
        {
            QuicClient client = new QuicClient();

            // Connect to peer (Server)
            QuicConnection connection = client.Connect(ip, port);

            // Create a data stream
            return connection.CreateStream(streamType);
        }

        /// <summary>
        /// Stop the Quic server
        /// </summary>
        /// <returns></returns>
        public static bool QuicStopServer()
        {
            if (_quicServer != null)
            {
                _quicServer.Close();
                Logger.Info($"The Quic server has been successfully stopped");
            }
            return true;
        }

        /// <summary>
        /// Start a Tcp server
        /// </summary>
        /// <returns><c>true</c>, start successfully, <c>false</c> did not start successfully.</returns>
        /// <param name="OnConnectCallBack">The callback function triggered when there is a client connection. The value passed in is the only client Token generated by the server.</param>
        /// <param name="OnReceivedCallBack">The callback function is triggered when the client receives information. The value passed in is the client Token and the string message received by the client. The returned string message is directly returned to the client If it is null, do not reply.</param>
        /// <param name="ip">The ip bound to the server. (default: ALL INTERFACES)</param>
        /// <param name="port">The listening port. (default: 8081)</param>
        /// <param name="iPType">ip type v4 or v6.</param>
        /// <param name="listen">Allowed concurrent connections (default: 10000)</param>
        public static bool TcpStart(
            Action<string> OnConnectCallBack,
            Func<string, byte[], byte[]> OnReceivedCallBack,
            int port = 8081,
            IPType iPType = IPType.IPv4,
            string ip = "0.0.0.0",
            int listen = 10000)
        {
            try
            {
                _tcpServer = new TcpServer(ip, port, iPType, OnConnectCallBack, OnReceivedCallBack, listen);

                _tcpServer.RunServer();

                Logger.Info($"The TCP server has been successfully started on port: {port}");
                return true;
            }
            catch (Exception exception)
            {
                Logger.Exception(exception);

                return false;
            }
        }

        /// <summary>
        /// Stop Tcp server
        /// </summary>
        public static void TcpStop()
        {
            if (_tcpServer != null)
            {
                _tcpServer.ShutDownServer();
            }
        }

        /// <summary>
        /// Broadcast a message to all clients on Tcp
        /// </summary>
        /// <param name="message">Message.</param>
        public static void TcpBroadCast(byte[] message)
        {
            if (_tcpServer != null)
            {
                _tcpServer.BroadCastMessageToAllClients(message);
            }
        }

        /// <summary>
        /// Send a message to the specified client according to the client Token
        /// </summary>
        /// <returns><c>true</c>, sent successfully, <c>false</c> did not send successfully.</returns>
        /// <param name="clientToken">Client Token.</param>
        /// <param name="message">The message to be sent.</param>
        public static bool TcpSend(string clientToken, byte[] message)
        {
            if (Token.Instance[clientToken] == null)
            {
                return false;
            }

            // no socket or no connection
            if (Token.Instance[clientToken].SocketHandler == null || !Token.Instance[clientToken].SocketHandler.Connected)
            {
                return false;
            }

            Token.Instance[clientToken].SendDataToClient(message);

            return true;
        }

        /// <summary>
        /// Close a client node
        /// </summary>
        /// <param name="clientToken">Client Token.</param>
        public static void TcpCloseClient(string clientToken)
        {
            try
            {
                if (Token.Instance[clientToken] != null)
                {
                    if (Token.Instance[clientToken].SocketHandler != null)
                    {
                        Token.Instance[clientToken].SocketHandler.Close();
                    }
                }
            }
            finally
            {
                // Assign null to delete the client Token
                Token.Instance[clientToken] = null;
            }
        }

        /// <summary>
        /// handler
        /// </summary>
        private static TcpServer _tcpServer;

        /// <summary>
        /// Start a Tcp client
        /// </summary>
        /// <returns><c>true</c>, i started successfully, <c>false</c> did not start successfully.</returns>
        /// <param name="ip">The IP to connect.</param>
        /// <param name="port">Port.</param>
        /// <param name="OnReceived">The callback function when a message is received. The value passed in is the received message. The returned string is sent directly to the server.
        /// If null is returned, it will not be sent.</param>
        public static bool TcpStart(string ip, int port, Func<byte[], byte[]> OnReceived, IPType ipType = IPType.IPv4)
        {
            _tcpClient = new TcpClient(ip, port, OnReceived, ipType);
            return true;
        }

        /// <summary>
        /// Tcp client sends a message
        /// </summary>
        /// <param name="msg">Message.</param>
        public static void TcpSend(byte[] msg)
        {
            if (_tcpClient == null)
            {
                return;
            }

            _tcpClient.SendDataToServer(msg);
        }

        /// <summary>
        /// handler
        /// </summary>
        private static TcpClient _tcpClient;

        /// <summary>
        /// Start Udp server
        /// </summary>
        /// <returns><c>true</c>, start successfully, <c>false</c> did not start successfully.</returns>
        /// <param name="port">Port. (default: 8082)</param>
        /// <param name="ResponseCallBack"> The return method of the received message accepts the incoming Endpoint source and string message, returns the string message, and returns the string type. The return value is directly returned to the client. If it is null, no reply is made.</param >
        /// <param name="iPType">IPv4 or IPv6.</param>
        /// <param name="bufferSize">Buffer size.</param>
        public static bool UdpStart(Func<System.Net.EndPoint, byte[], byte[]> ResponseCallBack = null, int port = 8082, IPType iPType = IPType.IPv4, int bufferSize = 1024)
        {
            try
            {
                _udpHandler = new Udp(port, iPType, ResponseCallBack, bufferSize);

                _udpHandler.runServer();

                Logger.Info($"UDP server has been successfully started on port: {port}");
                return true;
            }
            catch (Exception e)
            {
                Logger.Exception(e, "UDP Start exception");

                return false;
            }
        }

        /// <summary>
        /// Stop Udp server
        /// </summary>
        public static void UdpStop()
        {
            if (_udpHandler != null)
            {
                _udpHandler.StopServer();
            }
        }

        /// <summary>
        /// Set the Udp server's fallback method
        /// </summary>
        /// <param name="responseCallBack"> The method of returning the received message. Accept the incoming Endpoint source and string message and return the string type return value to directly reply to the client. If it is null, no reply. </param>
        public static void UDPCallBack(Func<System.Net.EndPoint, byte[], byte[]> responseCallBack)
        {
            if (_udpHandler != null)
            {
                _udpHandler.ResponseCallback = responseCallBack;
            }
        }

        /// <summary>
        /// Send a Udp message
        /// </summary>
        /// <param name="ip">Ip.</param>
        /// <param name="port">Port.</param>
        /// <param name="message">Message.</param>
        public static void UdpSend(string ip, int port, byte[] message, IPType ipType = IPType.IPv4)
        {
            if (_udpHandler == null)
            {
                _udpHandler = new Udp(8083, ipType);
            }

            _udpHandler.SendTo(new System.Net.IPEndPoint(System.Net.IPAddress.Parse(ip), port), message);
        }

        /// <summary>
        /// Send a Udp message
        /// </summary>
        /// <param name="target">Target.</param>
        /// <param name="message">Message.</param>
        public static void UdpSend(System.Net.EndPoint target, byte[] message, IPType ipType = IPType.IPv4)
        {
            if (_udpHandler != null)
            {
                _udpHandler = new Udp(8083, ipType);
            }

            _udpHandler.SendTo(target, message);
        }

        /// <summary>
        /// Get Udp server handle
        /// </summary>
        /// <returns>The UDP server.</returns>
        public static Udp GetUDPServer()
        {
            return _udpHandler;
        }

        /// <summary>
        /// handler
        /// </summary>
        private static Udp _udpHandler;

        /// <summary>
        /// Character bit encoding type web
        /// </summary>
        public static Encoding GlobalEncoding = Encoding.UTF8;

        private static void Main(string[] args)
        {
            TcpStart((s) => { Console.WriteLine(s + " connected"); }, (id, b) => { Console.WriteLine(GlobalEncoding.GetString(b)); return b; }, 8081, IPType.IPv4);
            UdpStart((id, b) => { Console.WriteLine(GlobalEncoding.GetString(b)); return b; }, 8082, IPType.IPv4);

            for (int i = 0; i < 5000; i++)
            {
                TcpStart("127.0.0.1", 8081, (byte[] receivedStr) => { return new byte[0]; }, IPType.IPv4);
            }

            for (int i = 0; i < 5000; i++)
            {
                UdpSend("127.0.0.1", 8082, Encoding.ASCII.GetBytes("HOI"), IPType.IPv4);
            }
            Console.ReadLine();
        }
    }

    /// <summary>
    /// IP type enumeration
    /// </summary>
    public enum IPType : byte
    {
        IPv4,
        IPv6
    }
}