﻿using System;
using System.Collections.Generic;
using System.Net;
using System.Net.Sockets;

namespace EonaCat.Network
{
    // This file is part of the EonaCat project(s) which is released under the Apache License.
    // See the LICENSE file or go to https://EonaCat.com/License for full license details.

    public class TcpServer
    {
        private readonly List<string> ClientTokens = new List<string>();
        private readonly Token Token = Token.Instance;

        public void BroadCastMessageToAllClients(byte[] msg, string exclusive = "")
        {
            foreach (string item in ClientTokens)
            {
                if (Token[item] != null /*&& item!= exclusive*/)
                {
                    Token[item].SendDataToClient(msg);
                }
            }
        }

        public void SendMessageToClient(string clientToken, byte[] msg)
        {
            if (Token[clientToken] != null)
            {
                Token[clientToken].SendDataToClient(msg);
            }
        }

        public void RunServer()
        {
            IPAddress ipAddress;

            switch (transportProtocol)
            {
                case IPType.IPv4:
                {
                    ipAddress = IPAddress.Any;
                }
                break;

                case IPType.IPv6:
                {
                    ipAddress = IPAddress.IPv6Any;
                }
                break;

                default:
                {
                    NetworkHelper.Logger.Error("Tcp Server IPv4 or IPv6 Setting Error!\n");
                    return;
                }
            }

            if (this.ipAddress != "0.0.0.0")
            {
                try
                {
                    ipAddress = IPAddress.Parse(this.ipAddress);
                }
                catch (Exception e)
                {
                    NetworkHelper.Logger.Exception(e, "Tcp Server Wrong Ip");
                }
            }

            IPEndPoint ipEndpoint;

            try
            {
                ipEndpoint = new IPEndPoint(ipAddress, port);
            }
            catch (Exception e)
            {
                NetworkHelper.Logger.Exception(e, "Tcp Server Wrong Port");
                return;
            }

            switch (transportProtocol)
            {
                case IPType.IPv4:
                {
                    listener =
                    new Socket(
                        AddressFamily.InterNetwork,
                        SocketType.Stream,
                        ProtocolType.Tcp);
                }
                break;

                case IPType.IPv6:
                {
                    listener =
                    new Socket(
                        AddressFamily.InterNetworkV6,
                        SocketType.Stream,
                        ProtocolType.Tcp);
                }
                break;

                default:
                {
                    NetworkHelper.Logger.Error("Tcp Server Socket Initialize error (at IPv4 or IPv6)\n");
                    return;
                }
            }

            listener.Bind(ipEndpoint);

            listener.Listen(ConcurrencyVolumn);

            listener.BeginAccept(new AsyncCallback(AcceptCallBack), listener);
        }

        public void ShutDownServer()
        {
            listener.Shutdown(SocketShutdown.Both);
            listener.Close();

            foreach (string client in ClientTokens)
            {
                try
                {
                    Token[client].SocketHandler.Close();
                    Token[client].SocketHandler.Shutdown(SocketShutdown.Both);
                }
                catch { }
            }

            try
            {
                TearDown?.Invoke();
            }
            catch (Exception e)
            {
                NetworkHelper.Logger.Exception(e);
            }

            ClientTokens.Clear();

            Token.Clear();
        }

        public Action TearDown;

        /// <summary>
        /// Pass in the client Token
        /// </summary>
        public Action<string> OnClientConnected;

        /// <summary>
        /// Incoming client Token, message msg
        /// </summary>
        public Func<string, byte[], byte[]> OnClientReceived;

        private void AcceptCallBack(IAsyncResult ar)
        {
            Socket ListenerHandler = (Socket)ar.AsyncState;

            Socket ClientHandler = ListenerHandler.EndAccept(ar);

            ListenerHandler.BeginAccept(new AsyncCallback(AcceptCallBack), ListenerHandler);

            string token = Token.GenerateGuid();

            Token[token] = new TcpConnectedPeer(token, ClientHandler, OnClientReceived, this);

            try
            {
                OnClientConnected?.Invoke(Token[token].Token + $" #{Token.Count}");
            }
            catch (Exception e)
            {
                NetworkHelper.Logger.Exception(e, "Tcp Server AcceptCallBack:::::::");
            }
        }

        private readonly string ipAddress;
        private readonly int port;
        private readonly int ConcurrencyVolumn;
        private readonly IPType transportProtocol;
        private Socket listener;

        public TcpServer(
            string _IPAddress,
            int _port,
            IPType _transProtocol,
            Action<string> OnClientConnected,
            Func<string, byte[], byte[]> OnClientReceived,
            int _ConcurrencyVolumn)
        {
            ipAddress = _IPAddress;

            port = _port;

            ConcurrencyVolumn = _ConcurrencyVolumn;

            transportProtocol = _transProtocol;

            this.OnClientConnected = OnClientConnected;

            this.OnClientReceived = OnClientReceived;
        }
    }
}