﻿using System;
using System.Net;
using System.Net.Sockets;

namespace EonaCat.Network
{
    // This file is part of the EonaCat project(s) which is released under the Apache License.
    // See the LICENSE file or go to https://EonaCat.com/License for full license details.

    public class Udp
    {
        public Socket socketHandler { get; private set; }

        public Func<EndPoint, byte[], byte[]> ResponseCallback;
        private IPEndPoint InitEndPoint;
        private readonly IPType IPType;
        private readonly int bufferSize;
        private readonly int port;

        public Udp(int Port = 8085, IPType IpType = IPType.IPv4, Func<EndPoint, byte[], byte[]> ResponseCallBack = null, int bufferSize = 1024)
        {
            ResponseCallback = ResponseCallBack;

            IPType = IpType;

            port = Port;

            this.bufferSize = bufferSize;

            init();
        }

        public void runServer()
        {
            if (socketHandler == null)
            {
                NetworkHelper.Logger.Error("IP Type Error at UDPServer Init");
                return;
            }

            UdpPeer peerInit = new UdpPeer(socketHandler, bufferSize, IPType);

            socketHandler.BeginReceiveFrom
                         (peerInit.buffer, 0, bufferSize, SocketFlags.None,
                          ref peerInit.remoteEndPoint, new AsyncCallback(BeginResponseCallBack), peerInit);
        }

        public void StopServer()
        {
            socketHandler.Close();
            socketHandler = null;
        }

        public void SendTo(EndPoint endPoint, byte[] msg)
        {
            if (socketHandler == null)
            {
                init();
            }

            this.socketHandler.BeginSendTo
                (msg, 0, msg.Length, SocketFlags.None, endPoint, new AsyncCallback(BeginSendToCallBack), null);
        }

        private void BeginResponseCallBack(IAsyncResult ar)
        {
            UdpPeer peer = (UdpPeer)ar.AsyncState;

            int rev = peer.serverSocket.EndReceiveFrom(ar, ref peer.remoteEndPoint);

            if (rev > 0)
            {
                if (ResponseCallback != null)
                {
                    byte[] result = ResponseCallback(peer.remoteEndPoint, peer.buffer);

                    if (result != null && result.Length > 0)
                    {
                        SendTo(peer.remoteEndPoint, result);
                    }

                    peer.ResetBuffer();
                }

                socketHandler.BeginReceiveFrom
                             (peer.buffer, 0, bufferSize, SocketFlags.None,
                              ref peer.remoteEndPoint, new AsyncCallback(BeginResponseCallBack), peer);
            }
        }

        private void BeginSendToCallBack(IAsyncResult ar)
        {
            socketHandler.EndSendTo(ar);
        }

        private void init()
        {
            if (IPType == IPType.IPv4)
            {
                socketHandler = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
                InitEndPoint = new IPEndPoint(IPAddress.Any, port);
            }
            else if (IPType == IPType.IPv6)
            {
                socketHandler = new Socket(AddressFamily.InterNetworkV6, SocketType.Dgram, ProtocolType.Udp);
                InitEndPoint = new IPEndPoint(IPAddress.IPv6Any, port);
            }
            socketHandler.Bind(InitEndPoint);
        }
    }

    public class UdpPeer
    {
        public EndPoint remoteEndPoint;
        public byte[] buffer;
        public Socket serverSocket;
        private readonly int bufferSize;
        private readonly IPType iPType = IPType.IPv4;

        public UdpPeer(Socket _serverSocket, int bufferSize, IPType iPType)
        {
            serverSocket = _serverSocket;

            this.iPType = iPType;

            if (this.iPType == IPType.IPv4)
            {
                remoteEndPoint = new IPEndPoint(IPAddress.Any, 0);
            }
            else
            {
                remoteEndPoint = new IPEndPoint(IPAddress.IPv6Any, 0);
            }

            buffer = new byte[bufferSize];

            this.bufferSize = bufferSize;

            this.iPType = iPType;
        }

        public void ResetBuffer()
        {
            buffer = new byte[bufferSize];

            if (iPType == IPType.IPv4)
            {
                remoteEndPoint = new IPEndPoint(IPAddress.Any, 0);
            }
            else
            {
                remoteEndPoint = new IPEndPoint(IPAddress.IPv6Any, 0);
            }
        }
    }
}